[defined] testing [if]
: .control ( -- ) 1 ; \ ."  Control char " ;
: .punct   ( -- ) 2 ; \ ."  Punctuation " ;
: .digit   ( -- ) 3 ; \ ."  Digit " ;
: .upper   ( -- ) 4 ; \ ."  Upper case letter " ;
: .lower   ( -- ) 5 ; \ ."  Lower case letter " ;  
: .nochar  ( -- ) 6 ; \ ."  Not a character " ;  
[else]
: .control ( -- ) ; \ ."  Control char " ;
: .punct   ( -- ) ; \ ."  Punctuation " ;
: .digit   ( -- ) ; \ ."  Digit " ;
: .upper   ( -- ) ; \ ."  Upper case letter " ;
: .lower   ( -- ) ; \ ."  Lower case letter " ;  
: .nochar  ( -- ) ; \ ."  Not a character " ;
[then]

: between ( c a b -- bool ) 1+ WITHIN ;
: cond  ( char -- 0 char ) 0 SWAP ; 
: range ( flag1 char low hi -- flag2 char ) 2 PICK >R  between OR  R> ;
: equal ( flag1 char val -- flag2 ) OVER >R  = OR  R> ;
: when  ( flag char -- char ) POSTPONE TUCK POSTPONE AND POSTPONE OF ; IMMEDIATE

: cond-when ( char -- ) 
  case
    cond
           00 $1F  range
          $7F      equal  when  .control endof
    cond
          BL  '/'  range
          ':' '@'  range
          '[' '`'  range
          '{' '~'  range  when  .punct   endof
    cond  '0' '9'  range  when  .digit   endof
    cond  'A' 'Z'  range  when  .upper   endof
    cond  'a' 'z'  range  when  .lower   endof
          >r .nochar r>
  endcase ;

: if-exit ( char -- ) { c }
    c $7F =   c $20 <  OR IF  .control  EXIT  ENDIF
    c '0' '9' between     IF  .digit    EXIT  ENDIF
    c 'A' 'Z' between     IF  .upper    EXIT  ENDIF
    c 'a' 'z' between     IF  .lower    EXIT  ENDIF
    c BL  '/' between     IF  .punct    EXIT  ENDIF 
    c ':' '@' between     IF  .punct    EXIT  ENDIF 
    c '[' '`' between     IF  .punct    EXIT  ENDIF 
    c '{' '~' between     IF  .punct    ELSE  .nochar  ENDIF ;

: if-exit-nolocal ( char -- )
    dup $7F =   over $20 <  OR IF drop .control  EXIT  ENDIF
    dup '0' '9' between     IF drop .digit    EXIT  ENDIF
    dup 'A' 'Z' between     IF drop .upper    EXIT  ENDIF
    dup 'a' 'z' between     IF drop .lower    EXIT  ENDIF
    dup BL  '/' between     IF drop .punct    EXIT  ENDIF 
    dup ':' '@' between     IF drop .punct    EXIT  ENDIF 
    dup '[' '`' between     IF drop .punct    EXIT  ENDIF 
        '{' '~' between     IF      .punct    ELSE  .nochar  ENDIF ;

: if-lt-exit ( char -- )
    dup bl  u< IF drop .control  EXIT  ENDIF
    dup '0' u< IF drop .punct    EXIT  ENDIF 
    dup ':' u< IF drop .digit    EXIT  ENDIF
    dup 'A' u< IF drop .punct    EXIT  ENDIF 
    dup '[' u< IF drop .upper    EXIT  ENDIF
    dup 'a' u< IF drop .punct    EXIT  ENDIF 
    dup '{' u< IF drop .lower    EXIT  ENDIF
    dup $7f u< IF drop .punct    EXIT  ENDIF
        $80 u< if      .control  exit endif
    .nochar ;

: if-tree ( char -- )
    dup '[' u< IF
	dup ':' u< IF
	    dup bl  u< IF drop .control  EXIT  ENDIF
	    dup '0' u< IF drop .punct    EXIT  ENDIF
	    drop .digit    EXIT  ENDIF
	dup 'A' u< IF drop .punct    EXIT  ENDIF
	drop .upper    EXIT  ENDIF
    dup '{' u< IF
	dup 'a' u< IF drop .punct    EXIT  ENDIF 
	drop .lower    EXIT  ENDIF
    dup $7f u< IF drop .punct    EXIT  ENDIF
        $80 u< if      .control  exit endif
    .nochar ;

: if-range ( char -- )
    dup 'a' 'z' between IF drop .lower    EXIT  ENDIF
    dup 'A' 'Z' between IF drop .upper    EXIT  ENDIF
    dup '0' '9' between IF drop .digit    EXIT  ENDIF
    dup  bl '~' between IF drop .punct    EXIT  ENDIF
    dup $80 u<          IF drop .control  EXIT  ENDIF
    drop .nochar ;

: range-of ( comp: -- of-sys; run-time: n1 n2 n3 -- | n1 )
    postpone 2>r postpone dup postpone 2r> postpone 1+ postpone within
    postpone ?of postpone drop ; immediate

: range-of1 ( char -- )
   case
     '0' '9'  range-of  .digit   endof
     'A' 'Z'  range-of  .upper   endof
     'a' 'z'  range-of  .lower   endof
     bl '~'  range-of   .punct   endof
     00 $7F  range-of   .control endof
           >r .nochar r>
   endcase ;

: range-of-reordered ( char -- )
   case
     'a' 'z'  range-of  .lower   endof
     'A' 'Z'  range-of  .upper   endof
     '0' '9'  range-of  .digit   endof
     bl '~'  range-of   .punct   endof
     00 $7F  range-of   .control endof
           >r .nochar r>
   endcase ;

: cond-when-no-or ( char -- )
  case
    cond  '0' '9'  range  when  .digit   endof
    cond  'A' 'Z'  range  when  .upper   endof
    cond  'a' 'z'  range  when  .lower   endof
    cond   bl '~'  range  when  .punct   endof
    cond   00 $7F  range  when  .control endof
          >r .nochar r>
  endcase ;

create test1-table $80 cells allot

: range! ( xt u1 u2 -- xt )
  1+ swap ?do
     dup test1-table i cells + ! loop 
  drop ;

' .control  0 $7f range!
' .punct  $20 $7e range!
' .digit  '0' '9' range!
' .upper  'A' 'Z' range!
' .lower  'a' 'z' range!

: execute1 ( c -- )
  dup $7f u> if
    drop .nochar exit then
  cells test1-table + @ execute ;

: case1 ( compilation  -- case-sys ; run-time  -- ) \ core-ext
    0 ; immediate

: ?of1 ( compilation  -- of-sys ; run-time  f -- ) \ gforth
    >r POSTPONE if r> ; immediate

: of1 ( compilation  -- of-sys ; run-time x1 x2 -- |x1 ) \ core-ext
    \ !! the implementation does not match the stack effect
    postpone over postpone = postpone ?of postpone drop ; immediate

: endof1 ( compilation case-sys1 of-sys -- case-sys2 ; run-time  -- ) \ core-ext end-of
    >r postpone else r> 1+ ; immediate

: n-thens1 ( orig1 ... origu u -- )
    0 ?do postpone then loop ;

: endcase1 ( compilation case-sys -- ; run-time x -- ) \ core-ext end-case
    >r postpone drop r> n-thens1 ; immediate

: cond1 ( comp: -- 0 )
    0 ; immediate

: dup-between ( x1 x2 x3 -- x1 f )
    1+ 2 pick >r within r> swap ;

: range1 ( comp: ucase u1 -- orig ucase u2; run-time: char low hi -- char )
    2>r postpone dup-between postpone 0= postpone if 2r> 1+ ; immediate

: equal1 ( comp: ucase u1 -- orig ucase u2; run-time: char low hi -- char )
    2>r postpone over postpone <> postpone if 2r> 1+ ; immediate

: when1 ( comp: orig1 ... origu u -- orig0 )
    swap >r >r postpone ahead
    r> 0 ?do
	1 cs-roll postpone then
    loop
    r> postpone drop ; immediate

: cond-when-shortcircuit ( char -- ) 
  case1
    cond1
           00 $1F  range1
          $7F      equal1  when1  .control endof1
    cond1
          BL  '/'  range1
          ':' '@'  range1
          '[' '`'  range1
          '{' '~'  range1  when1  .punct   endof1
    cond1 '0' '9'  range1  when1  .digit   endof1
    cond1 'A' 'Z'  range1  when1  .upper   endof1
    cond1 'a' 'z'  range1  when1  .lower   endof1
          >r .nochar r>
  endcase1 ;

: slurp-file ( c-addr1 u1 -- c-addr2 u2 ) \ gforth
    r/o bin open-file throw >r
    r@ file-size throw abort" file too large"
    dup allocate throw swap
    2dup r@ read-file throw over <> abort" could not read whole file"
    r> close-file throw ;

100000000 value n

: bench ( xt "name" -- ) { xt }
    parse-name slurp-file
    n over / 0 ?do
	2dup over + swap ?do
	    i c@ xt execute
	loop
    loop 2drop ;

[defined] testing [if]
: test { xt -- }
    13  xt execute 1 <> 1 and throw
    $7f xt execute 1 <> 1 and throw
    bl  xt execute 2 <> 1 and throw
    ':' xt execute 2 <> 1 and throw
    '[' xt execute 2 <> 1 and throw
    '{' xt execute 2 <> 1 and throw
    '0' xt execute 3 <> 1 and throw
    'A' xt execute 4 <> 1 and throw
    'a' xt execute 5 <> 1 and throw
    $80 xt execute 6 <> 1 and throw
;

.s
' cond-when test .s
' if-exit   test .s
' if-exit-nolocal test .s
' if-lt-exit test .s    
' if-tree test .s
' if-range test .s
' range-of1 test .s
' range-of-reordered test .s
' cond-when-no-or test .s
' execute1  test .s
' cond-when-shortcircuit test .s
[then]
